#pragma once

#include "lang_cxx.hpp"
#include "pkgconfig.hpp"

namespace {

	class gcc final : public compiler_cxx {
		fs::path gcc_exe, gxx_exe;
		pkgconfig pkgm;
		temp_file td{"runcpp-gcc", true};

		bool is_temp_file(string_view p) {
			const auto &r = td.name().native();
			if (p.size() < r.size())
				return false;
			if (p.substr(0, r.size()) != r)
				return false;
			if (p.size() > r.size() && p[r.size()] != '/')
				return false;
			return true;
		}

		void writefile(const fs::path &p, const auto &src) {
			auto f = unixfd::open(p.c_str(), O_WRONLY | O_CREAT | O_NOCTTY, 0666);
			for (const auto &l: src)
				f.writeall(l.data(), l.size());
		}

		vector<string_view> pkgopts(const map<string, set<string>> &lst) {
			set<string_view> pkgs;
			for (const auto &[pkg, _]: lst)
				pkgs.emplace(pkg);
			if (pkgs.empty())
				return {};
			return pkgm.get_options(pkgs);
		}

		unique_ptr<child_process>
			child_gcc(const compile_input &input, string_view infile, vector<string_view> extra) {
				vector<string_view> cmd;
				list<string> cmdbuf;

				auto _str = [&](string s) -> string_view {
					return cmdbuf.emplace_back(s);
				};

				auto push_1 = overloaded{
					[&](string_view s) {
						cmd.push_back(s);
					},
						[&](const char *s) {
							cmd.push_back(s);
						},
						[&](string s) {
							cmd.push_back(_str(s));
						},
						[&](vector<string_view> s) {
							for (auto e: s)
								cmd.emplace_back(e);
						},
				};
				auto push = [&](auto &&a1, auto &&...args) {
					if constexpr (sizeof...(args))
						push_1(vector<string_view>{ a1, args... });
					else
						push_1(std::forward<decltype(a1)>(a1));
				};

				const language *l = &lang();
				if (l == &lang_c)
					push(gcc_exe.native(), "-x", "c");
				else if (l == &lang_cxx)
					push(gxx_exe.empty() ? gcc_exe.native() : gxx_exe.native(), "-x", "c++");
				else
					throw std::logic_error("");

				push(std::move(extra));

				if (input.opt.std)
					push("-std="s + *input.opt.std);
				if (input.opt.optimization)
					push("-O"s + *input.opt.optimization);
				for (const auto &[n, v]: input.opt.macros)
					if (v)
						push("-D"s + n + "=" + *v);
					else
						push("-D"s + n);
				for (const auto &v: input.opt.includes)
					push("-I", v.native());
				for (const auto &v: input.opt.libpaths)
					push("-L"s + v.native());

				for (const auto &v: input.opt.lib[0])
					push("-l"s + v);
				push(pkgopts(input.opt.pkg[0]));

				push(infile.empty() ? "-"sv : infile);

				for (const auto &v: input.opt.lib[1])
					push("-l"s + v);
				push(pkgopts(input.opt.pkg[1]));

				if (verbose()) {
					auto s = "spawning"s;
					for (const auto &f: cmd) {
						s.append(" '");
						s.append(f);
						s.append("'");
					}
					verbose(s);
				}

				auto p = make_unique<child_process>(cmd);

				p->env("TMPDIR", td.name().native());

				return p;
			}

		void parse_cpp_mf(const fs::path &depfile, deplist &depl) {
			/* parse output of -MF, not makefile
			 * cannot handle \n
			 */
			string content;
			{
				auto f = unixfd::open(depfile.c_str(), O_RDONLY | O_NOCTTY);
				f.readall(content);
			}
			auto ci = content.begin();

			enum {
				blank,
				blankbackslash,
				word,
				backslash,
				dollar,
			} lstate = blank;
			string lbuf;
			unsigned bslashes = 0;

			enum {
				initial,
				prereqs,
			} pstate = initial;

			for (;;) {
				vector<string> token;
				bool quit = false;

				if (ci == content.end()) {
					switch (lstate) {
					case backslash:
						lbuf.append(bslashes, '\\');
						[[fallthrough]];
					case word:
						token.emplace_back(std::move(lbuf));
						[[fallthrough]];
					case blank:
						token.emplace_back("\n");
						break;
					default:
						throw runtime_error("invalid gcc deps file");
					}
					quit = true;
				} else {
					char c = *(ci++);
					switch (lstate) {
					case blank:
						switch (c) {
						case '\n':
							token.emplace_back("\n");
							[[fallthrough]];
						case ' ':
						case '\t':
							break;
						default:
							lbuf.clear();
							if (c == '\\') {
								lstate = blankbackslash;
								bslashes = 1;
							} else {
								lstate = word;
								lbuf.push_back(c);
							}
						}
						break;
					case word:
						switch (c) {
						case ' ':
						case '\t':
						case '\n':
							lstate = blank;
							token.emplace_back(std::move(lbuf));
							if (c == '\n')
								token.emplace_back("\n");
							break;
						case '\\':
							lstate = backslash;
							bslashes = 1;
							break;
						case '$':
							lstate = dollar;
							break;
						default:
							lbuf.push_back(c);
						}
						break;
					case blankbackslash:
						if (c == '\n') {
							lstate = blank;
							break;
						}
						[[fallthrough]];
					case backslash:
						switch (c) {
						case '\\':
							bslashes += 1;
							break;
						case ' ':
						case '\t':
							lbuf.append(bslashes / 2, '\\');
							if (bslashes % 2 == 1) {
								lbuf.push_back(c);
								lstate = word;
							} else {
								lstate = blank;
								token.emplace_back(std::move(lbuf));
							}
							break;
						case '\n':
							lbuf.append(bslashes - 1, '\\');
							break;
						case '#':
							bslashes -= 1;
							[[fallthrough]];
						default:
							lbuf.append(bslashes, '\\');
							lstate = word;
							lbuf.push_back(c);
						}
						break;
					case dollar:
						if (c != '$')
							throw runtime_error("invalid gcc deps file");
						lbuf.push_back(c);
						lstate = word;
						break;
					}
				}

				for (auto &t: token) {
					switch (pstate) {
					case initial:
						if (t.ends_with(":"))
							pstate = prereqs;
						break;
					case prereqs:
						if (t == "\n") {
							pstate = initial;
							break;
						}
						if (!is_temp_file(t))
							depl.add_file(t);
						break;
					}
				}

				if (quit)
					break;
			}
		}

		void parse_ld_trace(string_view out, deplist &r) {
			for (auto l: out | ranges::views::split('\n') |
					ranges::views::transform([](const auto &e) { return string_view(e.begin(), e.end()); }))
				if (l.starts_with("/") && !is_temp_file(l))
					r.add_file(l);
		}

	public:
		gcc(const language &l) :
			compiler_cxx(l),
			gcc_exe(child_process::find_executable("gcc")),
			gxx_exe(child_process::find_executable("g++")) {
				if (gcc_exe.empty())
					throw runtime_error("gcc not found");
			}

		string_view name() const noexcept override {
			return "GCC";
		}

		bool pkg_found(string_view p) override {
			return pkgm.pkg_found(p);
		}

		void compile(const compile_input &input, deplist &deps) override {
			auto infile = td.name() / "infile";
			writefile(infile, input.src);

			auto cppdepfile = td.name() / "cppdepfile";

			auto p = child_gcc(input, infile.native(), {
				"-MD", "-MF", cppdepfile.native(),
					"-fpic", "-fPIC",
					"-shared", "-Wl,--trace", "-rdynamic",
					"-o", input.output_path,
			});
			p->detach_stdin();
			auto outpipe = p->pipe(STDOUT_FILENO, 1);

			p->spawn();

			string deplibs;
			outpipe.readall(deplibs);

			if (p->wait() != 0)
				throw runtime_error("gcc failed");

			parse_cpp_mf(cppdepfile, deps);
			parse_ld_trace(deplibs, deps);
		}
	};

	static register_compiler<gcc> gcc_c{lang_c, 0};
	static register_compiler<gcc> gcc_cxx{lang_cxx, 0};

}
