#pragma once

#include "language.hpp"

namespace {

	class compiler_cxx : public compiler {
	public:
		using compiler::compiler;

		struct options {
			optional<string> std;
			optional<string> optimization;
			map<string, optional<string>> macros;
			set<fs::path> includes;
			set<fs::path> libpaths;
			array<vector<string>, 2> lib;
			array<map<string, set<string>>, 2> pkg, optpkg;

			auto tie() const noexcept {
				return std::tie(std, optimization, macros, includes, libpaths, lib, pkg, optpkg);
			}

			auto operator<=>(const auto &o) const noexcept {
				return tie() <=> o.tie();
			}
			bool operator==(const auto &o) const noexcept {
				return operator<=>(o) == 0;
			}

			bool parse_args(auto &&args_) {
				if (verbose()) {
					auto s = "found embedded options"s;
					for (const auto &a: args_) {
						s.append(" '");
						s.append(a);
						s.append("'");
					}
					verbose(s);
				}

				auto args = normalized_args(std::forward<decltype(args_)>(args_), {
					{"-std", true},
						{"-O", true},
						{"-D", true},
						{"-I", true},
						{"-L", true},
						{"-l", true},
						{"-p", true},
						{"-P", true},
						{"-@", false},
				});
				if (!args.positional.empty())
					throw runtime_error("unexpected argument '"s.append(args.positional.front()).append("'"));

				bool updated = false;

				for (const auto &a: args) {
					auto assign = overloaded{
						[&]<typename T>(optional<T> &dest) {
							optional<T> value{std::in_place, a[1]};
							if (dest == value)
								return;
							dest = std::move(value);
							updated = true;
						},
							[&]<typename T>(set<T> &dest, auto &&...args) {
								optional<T> value;
								if constexpr (sizeof...(args))
									value.emplace(std::forward<decltype(args)>(args)...);
								else
									value.emplace(a[1]);

								auto [_, inserted] = dest.emplace(std::move(*value));
								if (inserted)
									updated = true;
							},
							[&]<typename T>(vector<T> &dest) {
								optional<T> value{std::in_place, a[1]};
								dest.emplace_back(std::move(*value));
								updated = true;
							},
							[&]<typename T1, typename T2>(map<T1, T2> &dest) { /* for pkgs */
								string name{a[1]};
								string macro;
								if (auto p = name.find(':'); p != name.npos) {
									macro = string(name.substr(p + 1));
									name = name.substr(0, p);
								} else {
									macro = "FOUND_"s.append(name);
									for (char &c: macro)
										if (!(c == '_' ||
													(c >= '0' && c <= '9') ||
													(c >= 'A' && c <= 'Z') ||
													(c >= 'a' && c <= 'z')))
											c = '_';
								}

								auto i = dest.find(name);
								if (i == dest.end())
									std::tie(i, std::ignore) = dest.emplace(string(name), set<string>{});

								auto [_, inserted] = i->second.emplace(std::move(macro));
								if (inserted)
									updated = true;
							},
					};

					if (a[0] == "-std") {
						assign(std);
					} else if (a[0] == "-O") {
						assign(optimization);
					} else if (a[0] == "-D") {
						auto name = string(a[1]);
						optional<string_view> value;
						if (auto p = name.find('='); p != name.npos) {
							value = name.substr(p + 1);
							name = name.substr(0, p);
						}
						if (auto i = macros.find(name); i == macros.end() || i->second != value)
							updated = true;
						macros[name] = value;
					} else if (a[0] == "-I") {
						assign(includes, fs::absolute(a[1]));
					} else if (a[0] == "-L") {
						assign(libpaths, fs::absolute(a[1]));
					} else if (a[0] == "-l") {
						assign(lib[1]);
					} else if (a[0] == "-p") {
						assign(optpkg[1]);
					} else if (a[0] == "-P") {
						assign(pkg[1]);
					} else if (a[0] == "-@") {
						if (!lib[0].empty() || !pkg[0].empty() || !optpkg[0].empty())
							throw runtime_error("-@ could be specified at most once");
						lib[0] = std::move(lib[1]);
						pkg[0] = std::move(pkg[1]);
						optpkg[0] = std::move(optpkg[1]);
						if (!lib[0].empty() || !pkg[0].empty() || !optpkg[0].empty())
							updated = true;
					}
				}

				return updated;
			}

			void finish(compiler_cxx &);
		};

		struct compile_input {
			const options &opt;
			const vector<string_view> &src;
			string_view output_path;
		};

		virtual bool pkg_found(string_view) = 0;
		virtual void compile(const compile_input &, deplist &) = 0;

	private:
		string loaded_src;

		string_view load_src(const fs::path &src) {
			{
				auto f = unixfd::open(src.c_str(), O_RDONLY | O_NOCTTY);
				f.readall(loaded_src);
			}

			return loaded_src;
		}

		vector<string_view> parse_arglist(string_view a) {
			auto r = a |
				ranges::views::split(' ') |
				ranges::views::transform([](auto e) { return string_view(e.begin(), e.end()); }) |
				ranges::views::filter([](auto e) { return !e.empty(); });
			return {r.begin(), r.end()};
		}

		tuple<vector<string_view>, options> load_src_and_opt(const fs::path &srcpath) {
			auto rawsrc = load_src(srcpath);
			vector<string_view> lines;

			while (!rawsrc.empty()) {
				auto p = rawsrc.find('\n');
				if (p == rawsrc.npos)
					p = rawsrc.size() - 1;
				lines.emplace_back(rawsrc.substr(0, p + 1));
				rawsrc = rawsrc.substr(p + 1);
			}

			if (!lines.empty() && lines.front().starts_with("#!"))
				lines.front() = "\n";

			std::regex pragma{
				R"(^\s*#\s*pragma\s+runcpp (.*)$)"
			};

			options opt;

			for (auto &l: lines)
				if (std::cmatch m; !l.empty() && std::regex_match(l.begin(), l.end() - 1, m, pragma)) {
					auto args = parse_arglist({m[1].first, m[1].second});
					opt.parse_args(args);
					l = "\n";
				}
			opt.finish(*this);

			return {lines, opt};
		}

	public:
		void compile_lib(const fs::path &out, const fs::path &srcpath, deplist &deps) {
			auto [src, opt] = load_src_and_opt(srcpath);

			compile({
				.opt = opt,
					.src = src,
					.output_path = out.native(),
			}, deps);
		}
	};

	void compiler_cxx::options::finish(compiler_cxx &c) {
		for (const auto &[optlst, lst]: ranges::views::zip(optpkg, pkg)) {
			for (const auto &[p, pkgmacros]: optlst) {
				if (c.pkg_found(p)) {
					lst[p] = pkgmacros;
				} else {
					for (const auto &m: pkgmacros) {
						auto [_, inserted] = macros.emplace(m, "0");
						if (!inserted)
							throw runtime_error("macro '"s.append(m).append("' already defined"));
					}
				}
			}

			for (const auto &[_, pkgmacros]: lst)
				for (const auto &m: pkgmacros) {
					auto [_, inserted] = macros.emplace(m, "1");
					if (!inserted)
						throw runtime_error("macro '"s.append(m).append("' already defined"));
				}
		}
	}

	class language_cxx final : public language {
	public:
		using language::language;

		void show_help(std::ostream &s) const override {
			s << endl
				<< "C/C++ options:" << endl
				<< "  Use pragmas to embed options:" << endl
				<< "    * #pragma runcpp <option> [<option> ...]" << endl
				<< "" << endl
				<< "  -std <name>          Select language standard." << endl
				<< "  -O <level>           Select compiler optimization." << endl
				<< "  -D <name>[=<value>]  Define macro." << endl
				<< "  -I <path>            Add include directory." << endl
				<< "  -L <path>            Add library search path." << endl
				<< "  -l <name>            Search and link library." << endl
				<< "  -p <name>[:<macro>]  Search and link package if it exists, also defines <macro> to 0 or 1." << endl
				<< "                       Default macro name is FOUND_<name> with [^_0-9A-Za-z] replace with '_'." << endl
				<< "  -P <name>[:<macro>]  Like -p but require that package must exist." << endl
				<< "  -@                   Link all libraries before this option before the source file." << endl
				;
		}

		bool handle_fileext(string_view ext) const override {
			if (name() == "c" && ext == ".c")
				return true;
			if (name() == "c++" && (
						ext == ".cpp" ||
						ext == ".c++" ||
						ext == ".cxx" ||
						ext == ".cc"))
				return true;
			return false;
		}

		bool handle_vimft(string_view ft) const override {
			if (name() == "c" && ft == "c")
				return true;
			if (name() == "c++" && ft == "cpp")
				return true;
			return false;
		}

		void compile_lib(unique_ptr<compiler> comp_,
				const fs::path &cachedir, const fs::path &out,
				const fs::path &src,
				deplist &deps) const override {
			auto &comp = dynamic_cast<compiler_cxx &>(*comp_);
			comp.compile_lib(out, src, deps);
		}
	};

	static language_cxx lang_cxx{"c++"}, lang_c{"c"};

}
