#pragma once

#include "std.hpp"
#include "utils.hpp"
#include "deplist.hpp"

namespace {

	class language;

	class compiler {
		const language &_lang;
	public:
		virtual ~compiler() = default;
		compiler(const language &l) : _lang(l) {}

		const language &lang() const noexcept {
			return _lang;
		}

		virtual string_view name() const noexcept = 0;
	};

	class language {
		using compiler_create_fn = function<unique_ptr<compiler>(const language &)>;

		static inline lazy<map<string_view, const language *>> languages;

		lazy<multimap<int, compiler_create_fn>> compilers;
		string_view _name;
	public:
		language(string_view name) : _name(name) {
			auto [_a, added] = languages->emplace(name, this);
			if (!added)
				throw std::invalid_argument("");
		}

		string_view name() const noexcept {
			return _name;
		}

		void add_compiler(compiler_create_fn create, int prio=0) {
			compilers->emplace(prio, std::move(create));
		}
		unique_ptr<compiler> create_compiler() const {
			for (const auto &[_, c]: *compilers) {
				auto r = c(*this);
				if (r)
					return r;
			}
			return nullptr;
		}
		fs::path build_lib(const fs::path &cachedir, const fs::path &src) const {
			auto dep = cachedir / depfile();
			auto lib = cachedir / libfile();
			auto depnew = dep;
			depnew += ".new";
			auto libnew = lib;
			libnew += ".new";

			deplist deps;
			try {
				deps.load(dep);
				if (!deps.is_outdated(lib)) {
					verbose(lib, " up to date");
					return lib;
				}
			} catch (system_error &) {}
			deps.clear();

			verbose("building ", lib);

			auto comp = create_compiler();
			if (!comp)
				throw runtime_error("no compilers available");

			compile_lib(std::move(comp), cachedir, libnew, src, deps);

			deps.add_file(src);
			deps.store(depnew);

			fs::rename(depnew, dep);
			fs::rename(libnew, lib);

			return lib;
		}

		static const language *find_language(string_view name) {
			if (auto i = languages->find(name); i != languages->end())
				return i->second;
			return nullptr;
		}
		static const map<string_view, const language *> &list_languages() {
			return *languages;
		}

		static const language *detect(const fs::path &src) {
			auto ext = src.extension().string();
			std::transform(ext.begin(), ext.end(), ext.begin(), [](char c) { return std::tolower(c); });
			if (!ext.empty())
				for (const auto &[_, l]: *languages)
					if (l->handle_fileext(ext))
						return l;

			char buf[0x1000];
			string_view content;
			{
				auto f = unixfd::open(src.c_str(), O_RDONLY | O_NOCTTY);
				auto rs = f.read(buf, 0x1000);
				content = string_view(buf, rs);
			}

			std::regex re{
				R"([ \t][vV]im?:[ \t]?(?:set[ \t]+)([^\n]+[ :])(?:ft|filetype)=([a-zA-Z0-9]+))"
			};
			if (std::cmatch r; regex_search(content.begin(), content.end(), r, re, std::regex_constants::match_not_eol)) {
				if (r[1].matched) {
					string_view vimft(r[1].first, r[1].second);
					for (const auto &[_, l]: *languages)
						if (l->handle_vimft(vimft))
							return l;
				}
			}

			return nullptr;
		}

		virtual void show_help(std::ostream &) const = 0;
		virtual bool handle_fileext(string_view ext) const { return false; }
		virtual bool handle_vimft(string_view ft) const { return false; }
		virtual fs::path depfile() const { return "deps.txt"; }
		virtual fs::path libfile() const { return "lib.so"; }
		virtual void compile_lib(unique_ptr<compiler> comp, const fs::path &cachedir, const fs::path &out, const fs::path &src, deplist &deps) const = 0;
	};

	template<std::derived_from<compiler> T>
		struct register_compiler {
			template<std::derived_from<language> L>
				register_compiler(L &lang, int prio, const auto &...args) {
					lang.add_compiler(
							[argt = std::make_tuple<decltype(args)...>(args...)]
							(const language &l) -> std::unique_ptr<compiler> {
								try {
									return std::apply([&l](auto &&...a) {
										return make_unique<T>(l, std::forward<decltype(a)>(a)...);
									}, argt);
								} catch (runtime_error &) {}
								return nullptr;
							},
							prio
							);
				}
		};


}
