#pragma once

#include "std.hpp"

namespace {
	static constexpr size_t stack_size =
#ifndef __GNUC__
		std::max<size_t>(MINSIGSTKSZ,
#endif
				0x4000
#ifndef __GNUC__
				)
#endif
		;

	template<size_t N>
		struct fixed_output : std::array<char, N> {
			using array_t = std::array<char, N>;
			using array_t::data;
			using array_t::size;

			size_t pos = 0;

			void reset() {
				pos = 0;
			}

			string_view view() {
				return {data(), data() + pos};
			}

			void write(string_view v) {
				size_t s = std::min(v.size(), size() - pos);
				std::copy(v.begin(), v.begin() + s, data() + pos);
				pos += s;
			}

			void write(std::integral auto v) {
				char *p = data() + pos;
				auto [ptr, ec] = std::to_chars(p, data() + size(), v);
				if (ptr == p) {
					write("?");
				} else {
					pos += ptr - p;
				}
			}
		};

	using crash_buffer = fixed_output<0x10000>;

	struct crash_sigs {
		struct sig {
			int num;
			string_view name;
		};

		static inline constinit const sig sigs[] = {
#define D(n) { SIG##n, "SIG" #n }
			D(ABRT),
			D(BUS),
			D(FPE),
			D(ILL),
			D(QUIT),
			D(SEGV),
			D(SYS),
			D(TRAP),
			D(XCPU),
			D(XFSZ),
#undef D
		};

		static auto signals() {
			return sigs | ranges::views::transform([](const sig &s) { return s.num; });
		}

		static string_view name(int sig) {
			for (const auto &s: sigs)
				if (s.num == sig)
					return s.name;
			static fixed_output<10> buf;
			buf.reset();
			buf.write(sig);
			return buf.view();
		}
	};

	void crash_dump_sig(int sig, siginfo_t *info) {
		static crash_buffer buf;
		buf.write("\nReceived ");
		buf.write(crash_sigs::name(sig));
		buf.write(", crashed.\n");
		(void) ::write(STDERR_FILENO, buf.data(), buf.view().size());
	}

	[[maybe_unused]] void crash_dump_exc() {
		string_view msg = "\nUncaught exception, crashed.\n";
		(void) ::write(STDERR_FILENO, msg.data(), msg.size());
	}

}

extern "C" {
	static void rtenv_handle_crash(int sig, siginfo_t *info, void *context) noexcept {
		crash_dump_sig(sig, info);
		::_exit(0x7f + sig);
	}
}

namespace {

	void rtenv_clear_env() {
		char **dst = environ, **src = environ;
		for (; *src != nullptr; src++) {
			if (!string_view(*src).starts_with("runcpp_"))
				*(dst++) = *src;
		}
		*dst = nullptr;
	}

	void rtenv_setup() {
		rtenv_clear_env();

		alignas(max_align_t) static char stack_data[stack_size];
		stack_t ss = {
			.ss_sp = stack_data,
			.ss_size = stack_size,
		};

		if (::sigaltstack(&ss, NULL) < 0)
			throw system_error(errno, system_category(), "sigaltstack");

		struct ::sigaction sa{};
		sa.sa_flags = SA_ONSTACK | SA_SIGINFO;
		sa.sa_sigaction = rtenv_handle_crash;
		::sigfillset(&sa.sa_mask);

		for (auto s: crash_sigs::signals())
			if (::sigaction(s, &sa, nullptr) < 0)
				throw system_error(errno, system_category(), "sigaction "s.append(crash_sigs::name(s)));
	}

	int rtenv_call(auto f) noexcept {
#if 0
		try {
#endif
			return f();
#if 0
		} catch (...) {
			crash_dump_exc();
			::_exit(0x7f + SIGABRT);
		}
#endif
	}
}
