#include "gcc.hpp"
#include "language.hpp"
#include "rtenv.hpp"
#include "std.hpp"

extern "C" {
	typedef int (*main_fn)(int argc, char **argv);
}

namespace {

	class runcpp_app {
		string_view progname;

		const language *lang = nullptr;
		char *raw_src;
		vector<char *> arguments;

		void show_help(std::ostream &os) {
			std::ostringstream ss;

			ss << "usage: " << progname << " [-v]";

			if (lang)
				ss << " -x " << lang->name();
			else
				ss << " [-x lang]";

			ss
				<< " [--] <source> [argument ...]" << endl
				<< "Compile, load and execute source program." << endl
				<< endl
				<< "Options:" << endl
				<< "  -v  Dump more info." << endl
				<< "Environment:" << endl
				<< "  RUNCPP_CACHE    Set path to cache." << endl
				<< "  RUNCPP_VERBOSE  Show more info." << endl;

			if (lang)
				lang->show_help(ss);

			os << ss.view();
		}

		void show_lang_list(std::ostream &os) {
			std::ostringstream ss;

			ss << "Available languages:" << endl;
			for (const auto &[n, l]: language::list_languages()) {
				ss << "  " << n;
				auto c = l->create_compiler();
				if (c)
					ss << " (" << c->name() << ")";
				else
					ss << " (no compiler available)";
				ss << endl;
			}

			os << ss.view();
		}

		fs::path get_cache_root() {
			if (const char *e = ::getenv("RUNCPP_CACHE"); e)
				return e;
			auto name = "runcpp-cache-"s.append(std::to_string(getuid()));
			if (const char *e = getenv("TMPDIR"); e)
				return fs::path(e) / name;
			if (fs::is_directory("/var/tmp") && ::access("/var/tmp", W_OK | X_OK) == 0)
				return fs::path("/var/tmp") / name;
			return fs::temp_directory_path() / name;
		}

		void *compile_and_load() {
			auto src = fs::canonical(raw_src);

			if (!lang) {
				lang = language::detect(src);
				if (!lang)
					throw runtime_error("cannot detect supported language");
			}

			auto cache_root = get_cache_root();

			if (::access(cache_root.c_str(), F_OK) != 0) {
				fs::create_directories(cache_root.parent_path());
				if (::mkdir(cache_root.c_str(), 0700) < 0)
					throw system_error(errno, system_category(), "mkdir");
			}

			auto cached_dir = cache_root / "so" / src.lexically_relative("/");

			fs::create_directories(cached_dir);

			lockfile lock{cached_dir / "lock"};
			lock.lock_exclusive();

			fs::path libpath = lang->build_lib(cached_dir, src);

			void *lib = ::dlopen(libpath.c_str(), RTLD_NOW | RTLD_GLOBAL);
			if (lib == nullptr) {
				if (const char *e = dlerror(); e)
					throw runtime_error(e);
				throw runtime_error("cannot load compiled binary");
			}

			return lib;
		}

	public:
		void parse_args(int argc, char **argv) {
			progname = argv[0];
			const char *e = ::getenv("runcpp_argv0");
			if (e)
				progname = e;
			try {
				auto args = normalized_args(span(argv + 1, argc - 1), {
					{ "-v", false },
						{ "-x", true },
				});
				for (const auto &a: args) {
					if (a[0] == "-v") {
						::setenv("RUNCPP_VERBOSE", "1", 1);
					} else if (a[0] == "-x") {
						lang = language::find_language(a[1]);
						if (!lang) {
							show_lang_list(cerr);
							throw runtime_error("invalid language '"s.append(a[1]).append("'"));
						}
					}
				}
				if (args.positional.size() < 1)
					throw runtime_error("source file not specified");
				raw_src = const_cast<char *>(args.positional.front().data());
				auto srcargs = args.positional |
					ranges::views::drop(1) |
					ranges::views::transform([](auto e) { return const_cast<char *>(e.data()); });
				arguments = vector(srcargs.begin(), srcargs.end());
			} catch (runtime_error &) {
				show_help(cerr);
				cerr << endl;
				throw;
			}
		}

		int main() {
			void *lib = compile_and_load();

			auto *libmain = reinterpret_cast<main_fn>(::dlsym(lib, "main"));
			if (libmain == nullptr) {
				if (const char *e = dlerror(); e)
					throw runtime_error(e);
				throw runtime_error("cannot load main");
			}

			vector<char *> args = { raw_src };
			for (char *a: arguments)
				args.push_back(a);
			args.push_back(nullptr);

			rtenv_setup();

			return rtenv_call([libmain, &args] {
				return libmain(args.size() - 1, args.data());
			});
		}
	};

}

int main(int argc, char **argv) {
#ifdef NDEBUG
	try {
#endif
		runcpp_app app;
		app.parse_args(argc, argv);
		return app.main();
#ifdef NDEBUG
	} catch (std::exception &e) {
		cerr << "fatal error: " << e.what() << endl;
		return 127;
	}
#endif
}
