#pragma once

#include "std.hpp"

extern "C" {
	extern char **environ;
}

namespace {

	template<typename T>
		class lazy {
			mutable optional<T> data;
		public:
			constexpr lazy() = default;

			const T *operator->() const {
				if (!data)
					data.emplace();
				return &data.value();
			}
			T *operator->() {
				const lazy *cthis = this;
				return const_cast<T *>(cthis->operator->());
			}
			const T &operator*() const { return *operator->(); }
			T &operator*() { return *operator->(); }
		};

	class normalized_args : public vector<vector<string_view>> {
		vector<string> buf;
	public:
		struct missing_argument_error : runtime_error {
			missing_argument_error(const auto &a) :
				runtime_error("option '"s.append(a).append("' expects argument")) {}
		};
		struct unknown_option_error : runtime_error {
			unknown_option_error(const auto &a) :
				runtime_error("unknown option '"s.append(a).append("'")) {}
		};
		struct unexpected_argument_error : runtime_error {
			unexpected_argument_error(const auto &a) :
				runtime_error("option '"s.append(a).append("' does not take arguments")) {}
		};

		vector<string_view> positional;

		normalized_args(const auto &args, const map<string_view, bool> &opts) {
			bool posonly = false;
			vector<string_view> cur_opt;

			for (const auto &a_: args) {
				auto a = string_view(a_);

				if (a == "--") {
					posonly = true;
				} else if (posonly || (cur_opt.empty() && (a == "-" || !a.starts_with("-")))) {
					positional.push_back(a);
				} else if (!cur_opt.empty()) {
					if (a != "-" && a.starts_with("-"))
						throw missing_argument_error(cur_opt.front());
					cur_opt.push_back(a);
					push_back(std::move(cur_opt));
				} else if (auto p = a.find('='); p != a.npos) {
					auto name = a.substr(0, p);
					auto i = opts.find(name);
					if (i == opts.end())
						throw unknown_option_error(name);
					if (!i->second)
						throw unexpected_argument_error(name);
					push_back({name, a.substr(p + 1)});
				} else if (auto i = opts.find(a); i != opts.end()) {
					if (i->second)
						cur_opt.push_back(a);
					else
						push_back({a});
				} else if (a[1] == '-') {
					throw unknown_option_error(a);
				} else {
					for (size_t p = 1; p < a.size(); p++) {
						auto &name = buf.emplace_back(2, '-');
						name[1] = a[p];
						auto i = opts.find(name);
						if (i == opts.end())
							throw unknown_option_error(name);
						if (!i->second) {
							push_back({name});
						} else {
							if (p + 1 < a.size()) {
								push_back({name, a.substr(p + 1)});
								break;
							} else {
								cur_opt.push_back(name);
							}
						}
					}
				}
			}
			if (!cur_opt.empty())
				throw missing_argument_error(cur_opt.front());
		}
	};

	bool verbose() {
		static int cached = 0;
		if (cached == 0) {
			if (const char *e = ::getenv("RUNCPP_VERBOSE"); e)
				cached = 1;
			else
				cached = -1;
		}
		return cached == 1;
	}

	void verbose(auto &&a1, auto &&...args) {
		if (!verbose())
			return;
		std::ostringstream s;
		((s << "[RUNCPP " << ::getpid() << "] " << std::forward<decltype(a1)>(a1))
		 << ... << std::forward<decltype(args)>(args)) << endl;
		cerr << s.view();
	}

	class unixfd {
		int fd = -1;

		static auto check_call_neg(auto &&msg, auto f, const auto &...args) {
			auto r = f(args...);
			if (r == -1)
				throw system_error(errno, system_category(), msg);
			return r;
		}

	public:
		unixfd(int fd=-1) noexcept : fd(fd) {}
		~unixfd() { close(); }
		void swap(unixfd &o) noexcept { std::swap(fd, o.fd); }
		unixfd(unixfd &&o) noexcept { swap(o); }
		unixfd &operator=(unixfd o) noexcept { swap(o); return *this; }
		operator int() const noexcept { return fd; }

		void close() noexcept {
			if (fd >= 0) {
				::close(fd);
				fd = -1;
			}
		}

		int fcntl(const auto &...args) { return check_call_neg("fcntl", ::fcntl, fd, args...); }
		ssize_t read(void *buf, size_t nbyte) { return check_call_neg("read", ::read, fd, buf, nbyte); }
		ssize_t write(const void *buf, size_t nbyte) { return check_call_neg("write", ::write, fd, buf, nbyte); }

		void blocking(bool b) {
			int f = fcntl(F_GETFL);
			if (b)
				f &= ~O_NONBLOCK;
			else
				f |= O_NONBLOCK;
			fcntl(F_SETFL, f);
		}

		void cloexec(bool b) {
			int f = fcntl(F_GETFD);
			if (b)
				f &= ~FD_CLOEXEC;
			else
				f |= FD_CLOEXEC;
			fcntl(F_SETFD, f);
		}

		void readall(auto &s) {
			s.clear();
			for (;;) {
				auto sz = s.size();
				size_t cap = std::min(static_cast<size_t>(1024 * 1024 * 16), std::max(static_cast<size_t>(0x1000), sz));
				s.resize(sz + cap);
				auto rs = read(reinterpret_cast<void *>(&s[sz]), cap);
				s.resize(sz + rs);
				if (rs == 0)
					break;
			}
		}

		size_t writeall(const void *buf_, size_t nbyte) {
			const char *buf = reinterpret_cast<const char *>(buf_);
			size_t written = 0;
			while (nbyte > 0) {
				auto ws = write(buf, nbyte);
				if (ws == 0)
					break;
				written += ws;
				buf += ws;
				nbyte -= ws;
			}
			return written;
		}

		static unixfd open(const auto &...args) { return check_call_neg("open", ::open, args...); }
	};

	struct lockfile {
		unixfd fd;

		lockfile(const fs::path &path) {
			fd = unixfd::open(path.c_str(), O_RDWR | O_CREAT | O_NOCTTY | O_CLOEXEC, 0600);
		}

		void setlkw(short t) {
			struct ::flock l = {
				.l_type = t,
				.l_whence = SEEK_SET,
				.l_start = 0,
				.l_len = 0,
			};
			fd.fcntl(F_SETLKW, &l);
		}

		void lock_shared() {
			setlkw(F_RDLCK);
		}
		void lock_exclusive() {
			setlkw(F_WRLCK);
		}
		void unlock() {
			setlkw(F_UNLCK);
		}
	};

	class child_process {
		unique_ptr<::posix_spawn_file_actions_t, decltype([](auto *p){
			::posix_spawn_file_actions_destroy(p);
			delete p;
		})> _file_acts;
		list<string> _strbuf;
		vector<char *> _argv;
		string_view _file;
		vector<char *> _envbuf;
		char **_env = environ;
		bool _wait;
		::pid_t _pid = 0;
		vector<unixfd> _childfd;
		optional<int> status;

		char *_str(string_view s) {
			return _strbuf.emplace_back(s).data();
		}

		void file_act_add(auto f, auto ...args) {
			if (!_file_acts) {
				auto p = make_unique<::posix_spawn_file_actions_t>();
				if (::posix_spawn_file_actions_init(p.get()))
					throw std::bad_alloc();
				_file_acts.reset(p.release());
			}

			int e = f(_file_acts.get(), args...);
			switch (e) {
			case ENOMEM:
				throw std::bad_alloc();
			case 0:
				break;
			default:
				throw system_error(e, system_category(), "posix_spawn_file_actions_add");
			}
		}

	public:
		template<typename A=vector<string_view>>
			child_process(const A &args_={}, bool wait_=true) : _wait(wait_) {
				if (std::begin(args_) != std::end(args_))
					args(args_);
			}

		~child_process() {
			if (_pid > 0 && _wait) try {
				wait();
			} catch (...) {}
		}

		template<typename A=vector<string_view>, typename T=string_view>
			void args(const A &args_, optional<T> file_={}) {
				_argv.clear();

				for (auto a: args_)
					_argv.push_back(_str(a));
				_argv.push_back(nullptr);

				if (file_)
					_file = *file_;
				else
					_file = _argv[0];
			}

		void env(char **e) { _env = e; }
		void env(string_view name, optional<string_view> value={}) {
			if (_env) {
				_envbuf.clear();
				for (char **e = _env; *e != nullptr; e++)
					_envbuf.push_back(*e);
				_env = nullptr;
			}

			auto i = _envbuf.begin();
			for (; i != _envbuf.end(); i++)
				if (strncmp(*i, name.data(), name.size()) == 0 && (*i)[name.size()] == '=')
					break;

			if (value) {
				char *c = _str(string(name).append("=").append(*value));
				if (i != _envbuf.end())
					*i = c;
				else
					_envbuf.push_back(c);
			} else {
				if (i != _envbuf.end())
					_envbuf.erase(i);
			}
		}

		void close(int fd) {
			file_act_add(::posix_spawn_file_actions_addclose, fd);
		}
		void open(int fd, const char *path, int flags, mode_t mode) {
			file_act_add(::posix_spawn_file_actions_addopen, fd, path, flags, mode);
		}
		void dup2(int fd, int nfd) {
			file_act_add(::posix_spawn_file_actions_adddup2, fd, nfd);
		}

		void detach_stdin() {
			open(STDIN_FILENO, "/dev/null", O_RDONLY | O_NOCTTY, 0);
		}
		void detach_stdout() {
			open(STDOUT_FILENO, "/dev/null", O_WRONLY | O_NOCTTY, 0);
		}
		void detach_stdio() {
			detach_stdin();
			detach_stdout();
		}

		unixfd pipe(int fd, int child_idx) {
			int f[2];
			if (::pipe(f) < 0)
				throw system_error(errno, system_category(), "pipe");
			dup2(f[child_idx], fd);
			close(f[1 - child_idx]);
			_childfd.push_back(f[child_idx]);
			return f[1 - child_idx];
		}

		static fs::path find_executable(string_view name) {
			if (name.contains('/')) {
				auto r = fs::absolute(name);
				if (fs::is_directory(r) || ::access(r.c_str(), X_OK) != 0)
					return {};
			}

			string_view pathenv;
			if (const char *e = ::getenv("PATH"); e)
				pathenv = e;
			else
				return {};

			for (auto d: pathenv | ranges::views::split(':')) {
				if (d.begin() == d.end() || d.front() != '/')
					continue;
				auto r = fs::path(string_view(d.begin(), d.end())) / name;
				if (!fs::is_directory(r) && ::access(r.c_str(), X_OK) == 0)
					return r;
			}

			return {};
		}

		::pid_t spawn() {
			auto filepath = find_executable(_file).string();
			if (filepath.empty())
				throw system_error(ENOENT, system_category(), "cannot find '"s.append(_file).append("'"));

			if (!_env) {
				_envbuf.push_back(nullptr);
				_env = _envbuf.data();
			}

			int e = ::posix_spawn(&_pid, filepath.data(), _file_acts.get(), nullptr, _argv.data(), _env);
			if (e < 0)
				throw system_error(e, system_category(), "posix_spawn");

			_childfd.clear();

			return _pid;
		}

		int wait(int &st) {
			if (!status) {
				if (_pid <= 0)
					spawn();
				status.emplace();
				if (::waitpid(_pid, &*status, 0) < 0)
					throw system_error(errno, system_category(), "waitpid");
			}
			st = *status;
			return WEXITSTATUS(*status);
		}

		int wait() {
			int st;
			return wait(st);
		}
	};

	template<typename ...Ts>
		struct overloaded : private Ts... {
			overloaded(Ts ...f) : Ts(f)... {}
			using Ts::operator()...;
		};

	class temp_file {
		fs::path _name;
		bool remove = false;

		string prepare_template(string_view name) {
			return (fs::temp_directory_path() / name).string().append("-XXXXXX");
		}

		void init_file(string_view name) {
			string buf = prepare_template(name);
			int f = mkstemp(buf.data());
			if (f < 0)
				throw system_error(errno, system_category(), "mkstemp");
			_name = buf;
			remove = true;
			::close(f);
		}

		void init_dir(string_view name) {
			string buf = prepare_template(name);
			if (mkdtemp(buf.data()) == nullptr)
				throw system_error(errno, system_category(), "mkdtemp");
			_name = buf;
			remove = true;
		}

	public:
		temp_file(string_view name, bool isdir=false) {
			if (isdir)
				init_dir(name);
			else
				init_file(name);
		}

		~temp_file() {
			if (remove) {
				std::error_code ec;
				fs::remove_all(name(), ec);
			}
		}

		const fs::path &name() const noexcept {
			return _name;
		}
	};
}
